local _,N = ...
local getRewards, markItem = N.getRewards, N.markItem
local f = "|cFEFED000%s|r |T%s:0|t %s"

----------------------------------------------
-- Adapted from Blizzard_TalkingHeadUI.lua
----------------------------------------------

local speed = 0.5

local function FadeinFrames(self)
  self:Enable()
  C_Timer.After(0.75 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
  self.BackgroundFrame.Fadein:Play()
end

local function FadeoutFrames(self)
  if self.MainFrame.TextSheen:GetAlpha() > 0 then
    self.MainFrame.HideSheen:Play()
  end
  self.MainFrame.Close:Play()
  self.TextFrame.Close:Play()
  self.BackgroundFrame.Close:Play()
  self:Disable()
end

local function Reset(self, text)
  self:StopAnimating()
  self.BackgroundFrame.TextBackground:SetAlpha(0.01)
  self.TextFrame.Text:SetAlpha(0.00)

  self.MainFrame:SetAlpha(1)
  self.TextFrame.Text:SetText(text)
end

local function Close(self)
  self:FadeoutFrames()
end

local function OnClick(self)
  self:GetParent():Close()
  self.choose(self.index)
end

local function OnEnter(self, ...)
  --self.MainFrame.HideSheen:Stop()
  self.MainFrame.ShowSheen:Play()
  if self.link then
    GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT", 10, 0)
    GameTooltip:SetHyperlink(self.link)
    GameTooltip:Show()
    return
  end
  if self.choose == SelectGossipOption then
    local overriden = _G["GossipTitleButton" .. self.index]
    if not overriden then return end
    local script = overriden:GetScript("OnEnter")
    if not script then return end
    overriden:SetPoint("TOPLEFT", self)
    script(overriden, ...)
  end
end
local function OnLeave(self)
  --/self.MainFrame.ShowSheen:Stop()
  self.MainFrame.HideSheen:Play()
  GameTooltip:FadeOut()
end

local function Display(self, new)
  self.MainFrame.TextSheen:SetAlpha(0)
  new = new or not self:IsShown()
  local text = self.TextFrame.TextVirtual:GetText()

  if text then
    if new then
      self:Reset(text)
      self:FadeinFrames()
    else
      self.TextFrame.Fadeout:Play()
      C_Timer.After(0.25 * speed, function()
        self.TextFrame.Text:SetText(text)
      end)
      C_Timer.After(0.5 * speed, function()
        self.TextFrame.Fadein:Play()
      end)
    end
  end
end

local anim = 0.025

local frame = Conversation.ChoicesFrame

function frame:Reveal(from,to)
  for i=1,to do
    local child = select(i, self:GetChildren())
    C_Timer.After((to - i)*anim, function()
      child:Display(i >= from)
    end)
  end
end
function frame:Unreveal(from,to)
  for i=from,to do
    local child = select(i, self:GetChildren())
    child:ClearAllPoints()
    C_Timer.After((to - i)*anim, function()
      child:Close()
    end)
  end
end

local function CompleteAndGet(choice)
  CompleteQuest()
  GetQuestReward(choice)
end

local function Zero() return 0 end

local isGossip = true

local suggestWeight, suggestion

local function suggest(weight, ind)
  if weight >= suggestWeight then
    suggestWeight = weight
    suggestion = ind
  end
end

-- Suggestion Weights
local W_INCOMPLETE = 1
local W_GOSSIP = {}
for i,v in ipairs{ "gossip", "vendor", "trainer", "workorder" } do W_GOSSIP[v] = 2 + i/100 end
local W_OTHER = 3
local W_NEW = 4
local W_COMPLETE = 5

function frame:Refresh(event)
  local inlineRewards = N.Handler.core.rewards
  self.auto = nil
  suggestWeight = 0
  suggestion = nil
  if event == "QUEST_PROGRESS" and IsQuestCompletable() then event = "QUEST_COMPLETE" end
  local conf = N.Handler.core
  local  perRow,      baseHeight,        width,            scale,            differentHeights
  = conf.perRow, conf.buttonHeight, conf.buttonWidth, conf.buttonScale, conf.differentHeights
  local pad = 6

  if event == "GOSSIP_SHOW" then
    isGossip = true
  elseif event == "QUEST_GREETING" then
    isGossip = false
  end

  local NumAvailable, NumActive, NumGossip, Available, Active, Gossip, ChooseAvailable, ChooseActive, ChooseGossip
  if isGossip then
    NumAvailable = GetNumGossipAvailableQuests
    NumActive = GetNumGossipActiveQuests
    NumGossip = GetNumGossipOptions
    Available = GetGossipAvailableQuests
    Active = GetGossipActiveQuests
    Gossip = GetGossipOptions
    ChooseAvailable = SelectGossipAvailableQuest
    ChooseActive = SelectGossipActiveQuest
    ChooseGossip = SelectGossipOption
  else
    NumAvailable = GetNumAvailableQuests
    NumActive = GetNumActiveQuests
    NumGossip = Zero
    Available = GetAvailableQuestInfo
    Active = GetActiveTitle
    Gossip = GetGossipOptions
    ChooseAvailable = SelectAvailableQuest
    ChooseActive = SelectActiveQuest
    ChooseGossip = SelectGossipOption
  end

  local available = NumAvailable()
  local active = NumActive()
  local gossip = NumGossip()
  local oldSum = self.sum or 0
  local sum

  local choices = GetNumQuestChoices()
  if event == "QUEST_DETAIL" then
    if QuestGetAutoAccept() then
      sum = 0
    elseif inlineRewards then
      sum = 2
    else
      sum = 3
    end
  elseif event == "QUEST_COMPLETE" then
    sum = max(1, choices)
  else
    sum = available + active + gossip
  end
  for i=self:GetNumChildren(),sum do
    local child = CreateFrame("Button", nil, self, "FloatingQuestsChoiceTemplate")
    --child.TextFrame.Text:SetFontObjectsToTry(SystemFont_Shadow_Large, SystemFont_Shadow_Med2, SystemFont_Shadow_Med1)
    child.FadeinFrames = FadeinFrames
    child.FadeoutFrames = FadeoutFrames
    child.Reset = Reset
    child.Close = Close
    child.Display = Display
    child:SetScript("OnClick", OnClick)
    child:SetScript("OnEnter", OnEnter)
    child:SetScript("OnLeave", OnLeave)

    for i=1,select("#", child:GetChildren()) do
      local childChild = select(i, child:GetChildren())
      for j=1,select("#", childChild:GetAnimationGroups()) do
        local animGroup = select(j, childChild:GetAnimationGroups())
        for k=1,select("#", animGroup:GetAnimations()) do
          local anim = select(k, animGroup:GetAnimations())
          anim:SetStartDelay(anim:GetStartDelay() * speed)
          anim:SetDuration(anim:GetDuration() * speed)
        end
      end
    end

    N.Handler:BuildChoice(child)
  end

  self:Unreveal(sum+1,oldSum)

  if sum == 0 then return end

  for i=1,sum do select(i, self:GetChildren()).link = nil end

  if event == "QUEST_DETAIL" then
    local decline, viewRewards, accept
    if inlineRewards then
      decline, accept = self:GetChildren()
    else
      local rewards = getRewards(true)
      decline, viewRewards, accept = self:GetChildren()
      if rewards then
        viewRewards.TextFrame.TextVirtual:SetText(format(f, 2, "Interface\\GossipFrame\\IncompleteQuestIcon", "Rewards"))
      else
        viewRewards.TextFrame.TextVirtual:SetText("|cFF7777772 No rewards|r") -- TODO: Localize
      end
      viewRewards.choose = rewards and function()
        self:GetParent():SetText(getRewards(true) .. "\n\n\n|cFEFED000" .. GetObjectiveText() .. "|r")
      end or nil
    end
    decline.TextFrame.TextVirtual:SetText(format(f, inlineRewards and 2 or 3, "Interface\\RaidFrame\\ReadyCheck-NotReady", "|cFFDD0000Decline|r"))
    decline.choose = DeclineQuest
    accept.TextFrame.TextVirtual:SetText(format(f, 1, "Interface\\RaidFrame\\ReadyCheck-Ready", "|cFF00DD00Accept|r"))
    accept.choose = AcceptQuest
    suggest(W_COMPLETE, accept)
  elseif event == "QUEST_COMPLETE" then
    if choices < 2 then
      local complete = self:GetChildren()
      complete.index = 1
      complete.choose = CompleteAndGet
      complete.TextFrame.TextVirtual:SetText(format(f, 1, "Interface\\RaidFrame\\ReadyCheck-Ready", "|cFF00DD00Complete|r"))
      suggest(W_COMPLETE, complete)
    else
      for i=1,choices do
        local complete = select(i, self:GetChildren())
        complete.index = 1+choices-i
        complete.choose = CompleteAndGet
        local name, texture, numItems, quality, isUsable = GetQuestItemInfo("choice", complete.index)
        name = markItem(name)
        local color = isUsable and select(4, GetItemQualityColor(quality)) or "ffff0000"
        local displayName = name
        if numItems > 1 then displayName = numItems .. " " .. displayName end
        displayName = "|c" .. color .. displayName .. "|r"
        complete.TextFrame.TextVirtual:SetText(format(f, complete.index, texture, displayName))
        local link = GetQuestItemLink("choice", complete.index)
        local weight = W_COMPLETE
        if link then
          local price = select(11, GetItemInfo(link))
          if price then
            weight = weight + price
          end
        end
        complete.link = link
        suggest(weight, complete)
      end
    end
  else
    local at = 0

    for i=gossip,1,-1 do
      at = at + 1
      local child = select(at, self:GetChildren())
      local name,kind = select(1 + 2*(gossip-i), Gossip())
      child.choose = ChooseGossip
      child.index = 1 + gossip - i
      local tex = "Interface\\GossipFrame\\" .. kind .. "GossipIcon"
      child.TextFrame.TextVirtual:SetText(format(f, 1+sum-at, tex, name))
      suggest(W_GOSSIP[kind] or W_OTHER, child)
    end

    for i=active,1,-1 do
      at = at + 1
      local child = select(at, self:GetChildren())
      local name,level,isTrivial,isComplete,isLegendary
      if isGossip then
        name,level,isTrivial,isComplete,isLegendary = select(1 + 6*(active-i), Active())
      else
        local ind = 1+active-i
        name, isComplete = GetActiveTitle(ind)
        level = GetActiveLevel(ind)
        isTrivial = IsActiveQuestTrivial(ind)
        isLegendary = IsActiveQuestLegendary(ind)
      end
      child.choose = ChooseActive
      child.index = 1 + active - i

      child:SetPoint("BOTTOMRIGHT", self, "TOP")

      local tex
      if not isComplete then
        tex = "Interface\\GossipFrame\\IncompleteQuestIcon"
      elseif isLegendary then
        tex = "Interface\\GossipFrame\\ActiveLegendaryQuestIcon"
      else
        tex = "Interface\\GossipFrame\\ActiveQuestIcon"
      end
      child.TextFrame.TextVirtual:SetText(format(f, 1+sum-at, tex, name))

      if isComplete then
        suggest(W_COMPLETE, child)
      else
        suggest(W_INCOMPLETE, child)
      end
    end

    for i=available,1,-1 do
      at = at + 1
      local child = select(at, self:GetChildren())
      local name,level,isTrivial,frequency,isRepeatable,isLegendary
      if isGossip then
        name,level,isTrivial,frequency,isRepeatable,isLegendary = select(1 + 7*(available-i), Available())
      else
        local ind = 1 + available - i
        name = GetAvailableTitle(ind)
        level = GetAvailableLevel(ind)
        isTrivial, frequency, isRepeatable, isLegendary = GetAvailableQuestInfo(ind)
      end
      child.choose = ChooseAvailable
      child.index = 1 + available - i

      local tex
      if isLegendary then
        tex = "Interface\\GossipFrame\\AvailableLegendaryQuestIcon"
      elseif frequency == LE_QUEST_FREQUENCY_DAILY or frequency == LE_QUEST_FREQUENCY_WEEKLY then
        tex = "Interface\\GossipFrame\\DailyQuestIcon"
      elseif isRepeatable then
        tex = "Interface\\GossipFrame\\DailyActiveQuestIcon"
      else
        tex = "Interface\\GossipFrame\\AvailableQuestIcon"
      end

      child.TextFrame.TextVirtual:SetText(format(f, 1+sum-at, tex, name))
      suggest(W_NEW, child)
    end
  end

  perRow = min(perRow, sum)

  local fullWidth = width * perRow - pad * (perRow - 1)

  local lastPer = sum % perRow
  local lastWidth = width * lastPer - pad * (lastPer - 1)

  local line = self:GetChildren()
  local perLine = 0
  local prev

  for i=1,sum do
    local lastRow = i + perRow - 1 > sum
    local child = select(i, self:GetChildren())
    child:SetWidth(width)
    child:SetScale(scale)
    local text, virtual = child.TextFrame.Text, child.TextFrame.TextVirtual
    text:SetText()
    if i == 1 then
      child.offset = 0
      child:SetPoint("BOTTOMRIGHT", self, "TOP", fullWidth/2, -5)
      line = child
    elseif (i-1) % perRow == 0 then
      child.offset = 0
      local off
      if lastRow then
        off = -(fullWidth - lastWidth)/2
      else
        off = 0
      end
      child:SetPoint("BOTTOMRIGHT", line, "TOPRIGHT", off, -4)
      line = child
    else
      child:SetPoint("BOTTOMRIGHT", prev, "BOTTOMLEFT", pad, 0)
      if not lastRow or not differentHeights then
        child:SetPoint("TOPRIGHT", prev, "TOPLEFT", pad, 0)
      end
    end
    child:SetHeight(baseHeight)
    prev = child
  end

  if suggestion then
    local text = suggestion.TextFrame.TextVirtual
    text:SetText("|cFEFE6000" .. text:GetText():sub(11))
  end

  self.auto = suggestion

  C_Timer.After(0.001, function()
    local line = self:GetChildren()
    local perLine = 0
    local prev
    for i=1,sum do
      local lastRow = i + perRow - 1 > sum
      local child = select(i, self:GetChildren())
      local text, virtual = child.TextFrame.Text, child.TextFrame.TextVirtual
      text:SetText()
      if i == 1 then
        line = child
      elseif (i-1) % perRow == 0 then
        local off
        if lastRow then
          off = -(fullWidth - lastWidth)/2
        else
          off = 0
        end
        child:SetPoint("BOTTOMRIGHT", line, "TOPRIGHT", off, -4)
        line = child
      end
      local textHeight = text:GetHeight()
      local lineHeight = text:GetLineHeight()
      virtual:SetHeight(textHeight)
      local offset = 0
      for i=1,5 do
        if not virtual:IsTruncated() then
          break
        end
        offset = lineHeight * i
        virtual:SetHeight(textHeight + offset)
      end
      if lastRow and differentHeights then
        child:SetHeight(max(child:GetHeight() + offset, baseHeight + offset))
      else
        line:SetHeight(max(line:GetHeight(), child:GetHeight() + offset, baseHeight + offset))
      end
      prev = child
    end
  self.sum = sum
  self:Reveal(oldSum + 1, sum)
    end)
  self:GetParent():SetTop(prev)
end

function frame:Close()
  self.auto = nil
  self:Unreveal(1, self.sum or 0)
  self.sum = 0
  self:GetParent():SetTop()
end